# Sample script to test generate plots for the Large-Scale Inference Lecture
#
# 2014 written by Ralf Herbrich

# Converts a factorized Gaussian belief into a multivariate Gaussian
convertBelief <- function(wBelief) {
  return(
    list(mu = c(wBelief[[1]]$mu, wBelief[[2]]$mu),
         Sigma = matrix(c(wBelief[[1]]$sigma2,0,0,wBelief[[2]]$sigma2),2))
  );
}

# Computes the density of the 2D Gaussian
dnorm2d <- function(x,y,mu,Sigma, f=NA) {
  SigmaInv = solve(Sigma);
  res = x;
  c = 1/(2*pi*sqrt(det(Sigma)));
  for (i in 1:length(x)) {
    d = c(x[i],y[i]) - mu;
    if (is.na(f)) {
      res[i] = c*exp(-0.5*(t(d)%*%SigmaInv%*%d));      
    } else {  
      res[i] = c*exp(-0.5*(t(d)%*%SigmaInv%*%d))*f(x[i],y[i]);      
    }
  }
  return(res);
}

# Computes the likelihood of a product factor with Gaussian noise
prod <- function(x,y,z=2,sigma=0.5) {
  res = 1/(sqrt(2*pi)*sigma)*exp(-(x*y - z)*(x*y -z)/(2*sigma*sigma));
  return(res);
}

# Computes the likelihood of a inner product factor with Gaussian noise
linear <- function(w1,w2,y=2,x1=1,x2=-1,beta=0.5) {
  res = 1/(sqrt(2*pi)*beta)*exp(-(w1*x1+w2*x2 - y)*(w1*x1+w2*x2 -y)/(2*beta*beta));
  return(res);
}

# Plots the contour plot of a multivariate Gaussian multiplied with an arbitrary function
plotBelief <- function(mvGaussian, f=NA, N=100, levels=NA, ...) {
  x = seq(from=mvGaussian$mu[1]-3*sqrt(mvGaussian$Sigma[1,1]),
          to=mvGaussian$mu[1]+3*sqrt(mvGaussian$Sigma[1,1]),
          length.out=N);
  y = seq(from=mvGaussian$mu[2]-3*sqrt(mvGaussian$Sigma[2,2]),
          to=mvGaussian$mu[2]+3*sqrt(mvGaussian$Sigma[2,2]),
          length.out=N);
  z = outer(x,y,FUN=dnorm2d, f=f, mu=mvGaussian$mu, Sigma=mvGaussian$Sigma);
  if (length(levels) == 1 && is.na(levels)) {
    levels = pretty(range(z,finite=TRUE), 10);    
  }
  contour(x,y,z, nlevels=10, levels=levels, ...);
  return(levels);
}

# Plots the contour plot of an arbitrary function
plotLikelihood <- function(mvGaussian, f=NA, N=100, levels=NA, ...) {
  x = seq(from=mvGaussian$mu[1]-3*sqrt(mvGaussian$Sigma[1,1]),
          to=mvGaussian$mu[1]+3*sqrt(mvGaussian$Sigma[1,1]),
          length.out=N);
  y = seq(from=mvGaussian$mu[2]-3*sqrt(mvGaussian$Sigma[2,2]),
          to=mvGaussian$mu[2]+3*sqrt(mvGaussian$Sigma[2,2]),
          length.out=N);
  z = outer(x,y,FUN=f);
  if (length(levels) == 1 && is.na(levels)) {
    levels = pretty(range(z,finite=TRUE), 10);    
  }
  contour(x,y,z, nlevels=10, levels=levels, ...);
  return(levels);
}

# 2D demo of the algorithm
demo2DProd <- function() {  
  wPrior = list(list(mu=0,sigma2=1),list(mu=0,sigma2=1));
  wEPPosterior = list(list(mu=0,sigma2=1.1),list(mu=0,sigma2=1.1));
  wVBPosterior = list(list(mu=sqrt(2),sigma2=0.4),list(mu=sqrt(2),sigma2=0.4));
  
  levels = plotBelief(convertBelief(wPrior));
  readline(prompt="Press <ENTER> for next plot");
  
  levels = plotLikelihood(convertBelief(wPrior),f=prod);
  readline(prompt="Press <ENTER> for next plot");

  levels = plotBelief(convertBelief(wPrior),f=prod);
  readline(prompt="Press <ENTER> for next plot");

  plotBelief(convertBelief(wEPPosterior),col="red",add=TRUE,levels=levels);
  readline(prompt="Press <ENTER> for next plot");
  
  levels = plotBelief(convertBelief(wPrior),f=prod);
  plotBelief(convertBelief(wVBPosterior),col="green",add=TRUE,levels=levels);
  readline(prompt="Press <ENTER> for next plot");
}

# 2D demo of the algorithm
demo2DLinear <- function() {  
  wPrior = list(list(mu=0,sigma2=1),list(mu=0,sigma2=1));
  
  levels = plotBelief(convertBelief(wPrior));
  readline(prompt="Press <ENTER> for next plot");
  
  levels = plotLikelihood(convertBelief(wPrior),f=linear);
  readline(prompt="Press <ENTER> for next plot");
  
  levels = plotBelief(convertBelief(wPrior),f=linear);
  readline(prompt="Press <ENTER> for next plot");  
}

demo2DProd()
demo2DLinear()

